"""KAILINss - dynamic inference by integrating transcriptome and lineage information"""

import cospar as cs
import numpy as np
import os

from . import workcatalogue
from . import preprocessing
from . import kl_sample
from . import omic_functions
from . import database
from .kl_settings import *
from . import test_others
from . import database
__version__ = "1.0.0"

